/*
 *
 *
 *    daylog.cc
 *
 *    Version 0.9(v)
 *
 *    C++ Source Code for minimal Unix program to read text lines from
 *    standard input, proces them, and then save the results to a text
 *    file, rotating that file daily, or as otherwise specified.
 *
 *    Copyright (C) 2002,2003 by Matt Roberts, KK5JY,
 *    All rights reserved.
 *
 *    TODO (wishlist):
 *        1.  Insert code to allow auto-zipping of files each day.
 *        2.  Put back the 'pipe-to-command' capability.
 *
 *    ---------------- L I C E N S E    S T A T E M E N T -------------------
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *    ----------------------- E N D    L I C E N S E ------------------------
 *
 *
 */


#include <iostream>
#include <fstream>
#include <string>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <time.h>

#include "dir.h"
#include "stamps.h"
#include "util.h"

#define MAJOR_VERSION (0)
#define MINOR_VERSION (9)
#define REVISION ('v')

#define DEFAULT_FLAGS (O_WRONLY | O_CREAT)
#define NUM_BUFFERS (2)
#define IO_BUFFER_SIZE (4096)
#define BUFFER_LIMIT (IO_BUFFER_SIZE * 2)


/*
 *
 *
 *   Configuration Data
 *
 *
 */

// set to false to disable 'append' flags
static bool append = true;

// skip empty lines
static bool skip_empty = false;

// skip nonprintable characters
static unsigned short skip_nprint = true;

// line length limit
static size_t line_limit = 256;

// set to 'gmtime' for GMT, or 'localtime' for local time
static struct tm *(*time_func)(const time_t *timep) = 0;

// holds the format of the filename
static std::string fn_format;

// holds the format of the director
static std::string dn_format;

// holds the name of the root directory of the logs
static std::string cd_path;

// holds the name of the header file to add to each new file
static std::string header_path;

// what to use for end-of-line (default: Unix-style)
static std::string end_of_line = "\n";

// read check interval (sec) and last time checked
static size_t read_check_interval = 10;
time_t last_rotation_check = 0;

// function to timestamp lines
static std::string (*stamp_func)(const char * const aptr) = 0;
// pointer to optional parameter to above function
const char *sf_data = NULL;

// debugging flag
unsigned short d_debug = 0;


/*
 *
 *
 *    global program data
 *
 *
 */

// holds the time of the last log rotation
static time_t last_rotate = 0;

// set to true to exit normally
static volatile bool stop = false;

// set to true to reopen the current file
static volatile bool hup = false;

// holds the path to the currently open logfile
static std::string pathname;

// buffer for I/O
static unsigned char io_buffer[IO_BUFFER_SIZE];

// buffer between read and write actions
static std::string buffer;


/*
 *
 *   local_max()
 *
 */

static int local_max (int a, int b) {
	return (a >= b) ? a : b;
}


/*
 *
 *   SAFE_FD_ISSET(int fd, const fdset *)
 *
 */

static int SAFE_FD_ISSET(int fd, const fd_set *s) {
	if (fd < 0)
		return 0;
	return FD_ISSET(fd, s);
}


/*
 *
 *   signal_handler(int)
 *
 */

static void signal_handler (int sig_num) {
	switch (sig_num) {
		case SIGHUP:
			hup = true;
			break;
		default:
			stop = true;
			break;
	}
}


/*
 *
 *   setup_signals(void)
 *
 */

static void setup_signals(void) {
	// enable signal handlers
	signal(SIGHUP,  signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGPIPE, SIG_IGN);
}


/*
 *
 *   process_buffer(dst, src, ct)
 *
 */

static void process_buffer (std::string &dst, const unsigned char *src, const size_t ct) {
	static size_t last_nl = 0;
	const unsigned char *ip = src;
	char ch;

	if (d_debug) {
		std::cerr
			<< "process_buffer() called on "
			<< ct << " input bytes" << std::endl;
	}

	if (!ip)
		return;

	if (ct <= 0)
		return;

	while (ip < (src + ct)) {
		ch = *ip++;

		switch (ch) {
			case 0:     // ALWAYS strip NULL characters
			case '\r':  // ALWAYS strip CR characters
				break;

			case '\n':
				// TODO: clean up this logic to make more clear
				if ((!skip_empty) || (last_nl > 0)) {
					if (stamp_func && (last_nl == 0)) {
						dst += stamp_func(sf_data);
						dst += ' ';
					}
					dst += end_of_line;
				}
				last_nl = 0;
				break;

			default:
				if ((!skip_nprint) || (isprint(ch))) {
					if (stamp_func && (last_nl == 0)) {
						dst += stamp_func(sf_data);
						dst += ' ';
					}
					++last_nl;
					dst += ch;
				}
				break;
		}

		// if the buffer is full, or our line is now too long, start a new line
		if ((line_limit > 0) && (last_nl >= line_limit)) {
			dst += end_of_line;
			last_nl = 0;
		}
	}

	if (d_debug) {
		std::cerr
			<< "process_buffer() returning: " << dst.size()
			<< " bytes in destination." << std::endl;
	}
}


/*
 *
 *   dying_wish(...) - try to get all data flushed out to disk upon fatal error
 *
 */

static void dying_wish(int fd, std::string &buf) {
	int count = 0;

	if (d_debug > 1)
		std::cerr << "DEBUG: dying_wish entered" << std::endl;

	if (buf.empty())
		return;

	if (*(buf.end() - 1) != '\n')
		buf += end_of_line;

	while (!buf.empty()) {
		// WRITE:
		count = write(fd, buf.c_str(), buf.size());

		if (count < 0) {
			if (errno != EINTR) {
				if (!buf.empty()) {
					std::cerr
							<< "daylog: could not flush all data to disk: " << buf.size()
							<< ((buf.size() == 1) ? " byte " : " bytes ")
							<< "remaining" << std::endl
							<< "daylog: possible data loss" << std::endl;
					return;
				}
			}
		} else if (count > 0) {
			if ((size_t)count <= buf.size())
				buf.erase(0, count);
			else
				buf.erase();
		}
	}
}


/*
 *
 *   exit_action(ret_val)
 *
 */

static void exit_action (int ret_val) {
	// clean up signal handling
	signal(SIGHUP,  SIG_DFL);
	signal(SIGTERM, SIG_DFL);

	// return to parent
	exit(ret_val);
}


/*
 *
 *   usage(...)
 *
 */

static void usage (std::ostream &out) {
	out << std::endl;
	out << "daylog Version " << MAJOR_VERSION << '.'
		<< MINOR_VERSION << '(' << REVISION << ')' << std::endl << std::endl;
	out << "usage: daylog -c <dir> [options]" << std::endl << std::endl;
	out << "\t-h           help (this message)" << std::endl;
	out << "\t-c <dir>     change to <dir> before logging" << std::endl << std::endl;

	out << "\t-a           append mode (default)" << std::endl;
	out << "\t-A           overwrite mode" << std::endl;
	out << "\t-m <mode>    artificially modify umask" << std::endl << std::endl;

	out << "\t-d <format>  directory format (strftime-compliant string)" << std::endl;
	out << "\t-f <format>  filename format (strftime-compliant string)" << std::endl;
	out << "\t-p <string>  prefix for filenames (default is \"log-\")" << std::endl;
	out << "\t-s <name>    service name" << std::endl;
	out << "\t-g           measure time using GMT" << std::endl;
	out << "\t-G           measure time using local timezone (default)" << std::endl << std::endl;

	out << "\t-e           skip empty lines" << std::endl;
	out << "\t-E           allow empty lines (default)" << std::endl;
	out << "\t-j           skip nonprintable characters (default)" << std::endl;
	out << "\t-J           allow nonprintable characters" << std::endl;
	out << "\t-H <file>    filename of header file to insert into new logs" << std::endl;
	out << "\t-D           use CRLF for line ends" << std::endl << std::endl;

	out << "\t-i <integer> timecheck cycle (default 10 seconds)" << std::endl;
	out << "\t-l <integer> set maximum line length" << std::endl;
	out << "\t-z           increase debugging level" << std::endl << std::endl;

	out << "\t-r           Do not timestamp lines (default)" << std::endl;
	out << "\t-R           Use RFC 2822-compliant timestamp" << std::endl;
	out << "\t-u           Use ctime(3)-style timestamp" << std::endl;
	out << "\t-U           Use external TAI64N timestamp" << std::endl;
	out << "\t-t           Use time(2)-style timestamp, one second precision" << std::endl;
	out << "\t-T           Like -t, but with millisecond precision" << std::endl << std::endl;

	out << "-p, -s and -f are mutually exclusive, as are -r, -R, -t, -u and -U, and" << std::endl;
	out << "any option that has a counter-option of opposite case.  If more than one" << std::endl;
	out << "option in any mutually exclusive group are given together, the last option" << std::endl;
	out << "given will be used." << std::endl;
	out << std::endl;

	out << "The '-c' option is mandatory." << std::endl;
	out << std::endl;
}


/*
 *
 *   parse_args(...)
 *
 */

static void parse_args (int argc, char **argv) {
	// local vars
	int ch;
	mode_t new_umask;
	const char optstring[] = "aADeEgGjJorRtTuUzc:d:f:H:p:i:l:s:v::h::z";
	const char default_fn_suffix[] = "%Y%m%d.txt";

	// default values
	append     = true;
	time_func  = localtime;
	fn_format.erase();
	dn_format.erase();
	cd_path.erase();

	// for each arg element
	while ((ch = getopt(argc, argv, optstring)) != (-1)) {
		switch (ch) {
			case ':':
			case '?':
				exit_action(111);
			case 'a':
				append = true;
				break;
			case 'A':
				append = false;
				break;
			case 'm':
				new_umask = (int)strtoul(optarg, NULL, 8);
				umask(new_umask);
				break;
			case 'c':
				cd_path = optarg;
				break;
			case 'd':
				dn_format = optarg;
				break;
			case 'f':
				fn_format = optarg;
				break;
			case 'p':
				fn_format = std::string(optarg) + default_fn_suffix;
				break;
			case 's':
				fn_format = std::string(optarg) + "-" + default_fn_suffix;
				break;
			case 'H':
				header_path = (optarg);
				break;
			case 'o':
				append = 0;
				break;
			case 'j':
				skip_nprint = true;
				break;
			case 'J':
				skip_nprint = false;
				break;
			case 'e':
				skip_empty = true;
				break;
			case 'E':
				skip_empty = false;
				break;
			case 'G':
				time_func = localtime;
				break;
			case 'g':
				time_func = gmtime;
				break;
			case 'i':
				read_check_interval = (int)strtoul(optarg, NULL, 10);
				break;
			case 'l':
				line_limit = (int)strtoul(optarg, NULL, 10);
				if (line_limit == 0) break;
				if (line_limit < 80) line_limit = 80;
				if (line_limit > 1024) line_limit = 1024;
				break;
			case 'r':
				stamp_func = NULL;
				sf_data = NULL;
				break;
			case 'R':
				stamp_func = rfc2822_stamp;
				break;
			case 't':
				stamp_func = time_stamp;
				break;
			case 'T':
				stamp_func = ftime_stamp;
				break;
			case 'u':
				stamp_func = ctime_stamp;
				break;
			case 'U':
				stamp_func = tai64n_stamp;
				break;
			case 'D':
				end_of_line = "\r\n";
				break;
			case 'h':
			case 'v':
				usage(std::cerr);
				exit_action(127);
			case 'z':
				++d_debug;
				break;
			default:
				break;
		}
	}

	if (cd_path.empty()) {
		//std::cerr << "daylog: Must specify directory with -c" << std::endl;
		usage(std::cerr);
		exit_action(127);
	}

	if (fn_format.empty()) {
		fn_format = "log-";
		fn_format += default_fn_suffix;
	}

	if (dn_format.empty())
		dn_format = "%Y";
}


/*
 *
 *   write_header(fd)
 *
 */

void write_header(int fd) {
	// if there is no header file configured, just return
	if (header_path.empty())
		return;

	// if the file descriptor points to a file that already has
	// something in it, do nothing
	{
		off_t pos = lseek(fd, 0, SEEK_END);
		if (d_debug)
			std::cerr
				<< "DEBUG: Current file position: "
				<< (int)(pos) << std::endl;
		if (pos != (off_t)(0))
			return;
	}

	// if the file doesn't exist, issue warning, and do nothing else
	if (!fexist(header_path)) {
		std::cerr
			<< "daylog: Warning: header file \""
			<< header_path
			<< "\" does not exist, skipping." << std::endl;
		return;
	}

	// read the file into memory
	std::string data;
	std::ifstream infile(header_path.c_str());
	char ch;
	while (infile) {
		if (!infile.get(ch))
			break;
		// only allow non-NULL characters
		if (ch != 0)
			data += ch;
	}
	infile.close();

	if (d_debug)
		std::cerr
			<< "DEBUG: adding header of "
			<< data.size() << " bytes." << std::endl;

	// add the buffer to the beginning of the I/O buffer
	if (!data.empty())
		buffer.insert(0, data);

	if (d_debug)
		std::cerr
			<< "DEBUG: resulting I/O buffer now has "
			<< buffer.size() << " bytes." << std::endl;
}


/*
 *
 *   rotate_check(int*)
 *
 */

static void rotate_check(int *old_fd, time_t now = 0) {
	// hold the new pathname
	std::string newpath;

	// used to hold resultant file descriptor
	int new_fd = -1;

	// get the current time
	if (now == 0)
		now = time(NULL);

	// don't check too often:
	//    once per second is enough, especially on a busy logfile
	if ((!hup) && (last_rotation_check == now)) {
		if (d_debug)
			std::cerr << "Rotate check during same second, skipping." << std::endl;
		return;
	}

	// save time of this iteration
	last_rotation_check = now;

	// two tm structures for comparison
	struct tm time1, time2;

	// HUP: force logfile rotation, even if to same filename
	if (hup) {
		// reset HUP flag
		hup = false;

		// force rotation
		new_fd = -1;

		// DEBUG:
		if (d_debug)
			std::cerr
				<< "Rotate check requested as result of HUP: rotation forced"
				<< std::endl;
	}

	// if a file is already open, see if the time has changed sufficiently
	//    to warrant further checking
	if (new_fd >= 0) {
		// fill the tm structures
		time1 = *time_func(&last_rotate);
		time2 = *time_func(&now);

		// if time hasn't changed enough, go no further
		if (! times_differ(time1, time2)) {
			// DEBUG:
			if (d_debug)
				std::cerr
					<< "Rotate check requested but times_differ() returned false: skipping"
					<< std::endl;
			// return to caller
			return;
		}
	}

	//
	//  Now, build a new pathname, and then check to see if it is different
	//  from the current pathname.
	//

	// temp var for building strings
	char tmp[256];

	// start with empty path
	newpath.erase();

	// get current time
	now = last_rotate = time(NULL);

	// if a subdirectory structure is given, build it
	if (dn_format.size() > 0) {
		// construct directory name
		strftime(tmp, sizeof(tmp), dn_format.c_str(), time_func(&now));
		newpath = tmp;

		// make the new directory, and parents as needed
		mkdir_p(newpath, 0777);

		// copy the directory name into the newpath array
		newpath += '/';
	}

	// construct filename
	strftime(tmp, sizeof(tmp), fn_format.c_str(), time_func(&now));
	newpath += tmp;

	// sanity check
	if (newpath.empty() || (strlen(tmp) == 0)) {
		std::cerr
			<< "daylog: rotate_check(): internal error, empty file or pathname"
			<< std::endl;
		// try to get any remaining data into the old file
		if (*old_fd >= 0)
			dying_wish(*old_fd, buffer);
		exit_action(111);
	}

	// DEBUG:
	if (d_debug)
		std::cerr
			<< "Rotate check requested: newpath = " << newpath
			<< "; current pathname = " << pathname << std::endl;


	// if the newpath is the same as the old path, no need to rotate
	if (newpath == pathname) {
		// DEBUG:
		if (d_debug)
			std::cerr << "Pathnames do not differ, rotation skipped." << std::endl;

		// return to caller
		return;
	} else {
		// close the old file
		new_fd = -1;

		// DEBUG:
		if (d_debug)
			std::cerr << "Pathnames differ, rotation will be performed." << std::endl;
	}

	// if anything above asked to close the logfile (or if it was not
	//    already opened), open a new logfile
	if (new_fd < 0) {
		// holds current file I/O flags
		int flags;

		// set new pathname
		pathname = newpath;

		// compute the 'open' flags
		flags = DEFAULT_FLAGS;
		if (append)
			flags |= O_APPEND;

		// DEBUG: display filename
		if (d_debug)
			std::cerr << "Opening file: " << pathname << std::endl;

		// open the new file, and bail if it fails
		new_fd = open(pathname.c_str(), flags, 0666);
	}

	/*
	 *  last sanity check:
	 *     if the function wasn't able to open a new file, the program dies here
	 */

	if (new_fd < 0) {
		std::cerr
			<< "daylog: rotate_check(): could not open new logfile, errno = "
			<< errno << " (" << strerror(errno) << ")" << std::endl;
		// BAILOUT: if we didn't already have a file open, bail
		if (*old_fd < 0)
			exit_action(111);
		// return to caller
		return;
	}

	//
	//  SUCCESS!
	//

	// if we had an open file, close it
	if (*old_fd >= 0)
		close(*old_fd);

	// set the new value of the fd
	*old_fd = new_fd;

	// if we need to insert a header into the new file, do so
	if (! header_path.empty())
		write_header(new_fd);
}


/********************************************************************\
 *                                                                  *
 *                                                                  *
 *                                                                  *
 *                                                                  *
 *     main ()                                                      *
 *                                                                  *
 *                                                                  *
 *                                                                  *
 *                                                                  *
\********************************************************************/

int main (int argc, char **argv) {
	// for holding time values
	struct timeval timeout;

	// 'n' for 'select()'
	int n = 0;

	// used in 'select' call
	fd_set read_set;
	fd_set write_set;

	// input fd to read logfile data:
	//    start with 'stdin' for input; may change this later
	int input_fd = STDIN_FILENO;

	// set nonblocking on standard input
	//    ...(as if it matters when using select())
	fcntl(input_fd, F_SETFL, O_NONBLOCK);

	// output to open logfile:
	//    start with no output, later function will assign
	int file_fd = -1;

	// used for I/O
	ssize_t count = 0;

	// ARGS: load settings
	parse_args (argc, argv);

	// start with empty buffers
	io_buffer[0] = 0;
	buffer.erase();

	// SIGNALS: setup signal handling
	setup_signals();

	// CHDIR: change into desired directory
	if (chdir(cd_path) < 0) {
		std::cerr
			<< "daylog: could not change to directory \""
			<< cd_path << "\", errno = " << errno
			<< " (" << strerror(errno) << ")" << std::endl;
		// BAILOUT: can't proceed if we can't find the directory
		exit_action(111);
	}

	// OPEN: make sure we initially have a current open logfile
	rotate_check(&file_fd);

	/*
	 *
	 *   main loop
	 *
	 */

	while (!stop) {
		if (d_debug > 1) {
			std::cerr
				<< "Begin main loop: len(buffer) = "
				<< buffer.size() << ", input_fd = " << file_fd << std::endl;
		}

		// EOF: check for end-of-file on input
		if ((input_fd < 0) && (buffer.size() == 0))
			exit_action(0);

		// HUP: if we receive a HUP, force rotation of the logfile
		if (hup)
			rotate_check(&file_fd);
	
		// TIMEOUT: build timeout structure
		timeout.tv_sec = read_check_interval;
		timeout.tv_usec = 0;

		// SETUP: zero out the fd sets
		FD_ZERO(&read_set);
		FD_ZERO(&write_set);

		// start with no descriptors to watch
		n = 0;

		// READ EVENT: if we have input buffer space
		if ((input_fd >= 0) && (buffer.size() < BUFFER_LIMIT)) {
			FD_SET(input_fd, &read_set);
			n = local_max(n, input_fd);
		}

		// WRITE EVENT: if we have any complete lines waiting to be written, or if our buffer is full
		if (has_lines(buffer) || (buffer.size() >= BUFFER_LIMIT)) {
			FD_SET(file_fd, &write_set);
			n = local_max(n, file_fd);
		}

		// DEBUG:
		if (d_debug > 2) {
			std::cerr << std::endl << "****************************************************" << std::endl;
			if (SAFE_FD_ISSET(input_fd, &read_set)) std::cerr << "* SOURCE READ ENABLED" << std::endl;
			if (SAFE_FD_ISSET(file_fd, &write_set)) std::cerr << "* FILE WRITE ENABLED" << std::endl;
			std::cerr << "* Call select(" << (n + 1) << ", ...)" << std::endl;
		}

		// SELECT: wait for data
		count = select(n + 1, &read_set, &write_set, NULL, &timeout);

		// DEBUG:
		if (d_debug > 2) {
			std::cerr << "* " << count << " EVENTS RETURNED:" << std::endl;
			if (SAFE_FD_ISSET(input_fd, &read_set)) std::cerr << "* SOURCE READ EVENT" << std::endl;
			if (SAFE_FD_ISSET(file_fd, &write_set)) std::cerr << "* FILE WRITE EVENT" << std::endl;
			std::cerr << "****************************************************" << std::endl << std::endl;
		}

		// CHECK: check for errors
		if (count < 0) {
			if (stop) {
				dying_wish(file_fd, buffer);
				exit_action(0);
			}

			// if our error was simply an interrupted signal handler, try again
			if (errno == EAGAIN)
				continue;

			if ( ! hup) {
				std::cerr
					<< "daylog: select() returned errno = "
					<< errno << " (" << strerror(errno) << ")" << std::endl;
				dying_wish(file_fd, buffer);
				// BAILOUT: something bad happened, quit
				exit_action(111);
			}
		}

		// AGAIN: if select didn't have any events, try again
		if (count <= 0) {
			// ROTATE: rotate out the log, if needed
			if ((count == 0) || hup)
				rotate_check(&file_fd);

			// go around the loop again
			continue;
		}

		// EVENT: file write event
		if (SAFE_FD_ISSET(file_fd, &write_set)) {
			// move to new log, if needed
			rotate_check(&file_fd);

			// DEBUG: ...
			if (d_debug > 2)
				std::cerr << "FILE: calling write_line(): buffer size = " << buffer.size() << std::endl;

			// WRITE: write one line
			if (!write_line(file_fd, buffer, end_of_line)) {
				std::cerr
					<< "daylog: write returned errno = "
					<< errno << " (" << strerror(errno) << ")" << std::endl;
				dying_wish(file_fd, buffer);
				// BAILOUT: something bad happened, quit
				exit_action(111);
			}

			// DEBUG: ...
			if (d_debug > 2)
				std::cerr << "FILE: write_line() successful: buffer size = " << buffer.size() << std::endl;
		}

		// EVENT: source read event
		if (SAFE_FD_ISSET(input_fd, &read_set)) {
			// DEBUG: ...
			if (d_debug)
				std::cerr << "STDIN: calling read(): buffer size = " << buffer.size() << std::endl;

			// READ: read input from stdin
			count = ::read(input_fd, io_buffer, sizeof(io_buffer));

			// DEBUG: ...
			if (d_debug)
				std::cerr << "STDIN: read() returned: " << count << std::endl;

			// ERROR:
			if (count < 0) {
				if (errno != EAGAIN) {
				std::cerr
					<< "daylog: read(0) returned errno = "
					<< errno << " (" << strerror(errno) << ")" << std::endl;
					exit(111);
				}
			}

			// EOF: if we reach EOF, close the source
			if (count == 0) {
				close(input_fd);
				input_fd = (-1);
			}

			// FILTER: run the input through the filter
			process_buffer(buffer, io_buffer, count);
		}
	}

	//
	// EXIT: if a signal handler set the 'stop' var, we should exit
	//

	// BAILOUT: save data to file and quit
	dying_wish(file_fd, buffer);
	exit_action(0);

	// KLUDGE: this line to shut up gcc warnings
	return 0;
}

// end-of-file: daylog.cc
